from typing import Any, Union, Callable, Optional, cast, Literal, Type
from pydantic import Field

from .base import Event, MessageEvent


class OfficalAccountEvent(Event):
    """ 公众号 事件 """


class OfficalAccountMessageEvent(MessageEvent, OfficalAccountEvent):
    """ 公众号 用户消息事件 """
    msg_data_id: Optional[str] = Field(default=None, alias="MsgDataId")
    """ 消息的数据ID（消息如果来自文章时才有） `MsgDataId` """
    msg_article_id: Optional[str] = Field(default=None, alias="Idx")
    """ 消息的文章ID（消息如果来自文章时才有） `Idx` """


class SubscribeEvent(OfficalAccountEvent):
    """ 公众号 用户关注事件 """
    message_type: Literal["event"] = "event"
    event: Literal["subscribe"] = "subscribe"

    event_key: Optional[str] = Field(alias="EventKey")
    """ 带参数的公众号二维码，二维码的参数值 `EventKey` """
    ticket: Optional[str] = Field(alias="Ticket")
    """ 二维码的 ticket，可用来换取二维码图片 `Ticket` """


class UnSubscribeEvent(OfficalAccountEvent):
    """ 公众号 用户取消关注事件 """
    message_type: Literal["event"] = Field(alias="MsgType")
    event: Literal["unsubscribe"] = Field(alias="Event")


class MenuClickEvent(OfficalAccountEvent):
    """ 公众号 菜单点击事件 """
    message_type: Literal["event"] = Field(alias="MsgType")
    event: Literal["CLICK"] = Field(alias="Event")

    event_key: str = Field(alias="EventKey")
    """ 事件KEY值，与自定义菜单接口中KEY值对应 `EventKey` """


class AuthorizationChangeEvent(OfficalAccountEvent):
    """ 授权用户信息变更事件 """
    message_type: Literal["event"] = Field(alias="MsgType")
    """ 消息类型 `MsgType` """
    event: Literal["user_authorization_revoke"] = Field(alias="Event")
    """ 事件类型 `Event` """
    openid: str = Field(alias="OpenID")
    """ 用户 OpenID `OpenID """
    appid: str = Field(alias="AppID")
    """ 公众号/小程序 AppID `AppID` """
    revoke_info: str = Field(alias="RevokeInfo")
    """ 取消授权的数据类型 `RevokeInfo` """



from pydantic import Field, HttpUrl, BaseModel
from typing import Literal


class BotInfo(BaseModel):
    appid: str = Field()
    token: str = Field()  # 事件推送令牌
    secret: str = Field()  # 接口调用凭证
    enable_verify: bool = False  # 是否启用响应事件推送验证
    type: Literal["official", "miniprogram"] = Field(default="miniprogram")  # 机器人类型 小程序/公众号：miniprogram / official
    callback: HttpUrl = Field(default=None)  # 是否将事件推送转发到指定 URL


class Config(BaseModel):
    wxmp_bots: list[BotInfo] = Field(default_factory=list)



from typing import Union, Any, Optional, Type, TYPE_CHECKING, cast
from typing_extensions import override
import json
import time

from nonebot.message import handle_event
from nonebot.utils import logger_wrapper
from nonebot.adapters import Bot as BaseBot
from nonebot.drivers import Request, Response
from nonebot.exception import (
    ActionFailed,
    NetworkError,
    ApiNotAvailable,
)

from .event import *
from .config import BotInfo
from .message import Message, MessageSegment

if TYPE_CHECKING:
    from .adapter import Adapter


log = logger_wrapper("WXMP")


class Bot(BaseBot):

    @override
    def __init__(self, adapter: "Adapter", self_id: str, bot_info: BotInfo):
        super().__init__(adapter, self_id)
        self.bot_info: BotInfo = bot_info

        self._access_token: Optional[str] = None
        self._expires_in: Optional[int] = None

    @override
    async def send(
        self,
        event: Event,
        message: Union[str, Message, MessageSegment],
        **kwargs,
    ) -> Any:
        """ 发送消息 """
        if isinstance(event, MiniprogramMessageEvent) \
                or isinstance(event, OfficalAccountMessageEvent):
            return await self.send_custom_message(event.user_id, message)
        else:
            raise ApiNotAvailable()

    async def handle_event(self, event: Type[Event]):
        """ 处理事件 """
        await handle_event(self, event)

    async def _get_access_token(self) -> str:
        """ 获取微信公众平台的 access_token """
        now = time.time()
        if (self._expires_in or 0) > now:
            return self._access_token

        request = Request(
            method="POST",
            url="https://api.weixin.qq.com/cgi-bin/stable_token",
            json={
                "grant_type": "client_credential",
                "appid": self.bot_info.appid,
                "secret": self.bot_info.secret,
                "force_refresh": False,
            },
        )
        resp = await self.adapter.request(request)
        if resp.status_code != 200 or not resp.content:
            raise NetworkError(
                f"Get authorization failed with status code {resp.status_code}."
                " Please check your config."
            )
        res: dict = json.loads(resp.content)
        self._expires_in = now + res["expires_in"]
        self._access_token = res["access_token"]
        return self._access_token

    async def send_custom_message(self, user_id: str, message: Message):
        """ 发送 客服消息 """
        if isinstance(message, str):
            message = Message(MessageSegment.text(message))
        elif isinstance(message, MessageSegment):
            message = Message(message)
        elif not isinstance(message, Message):
            raise ValueError("Unsupported message type")

        for segment in message:
            segment = cast(MessageSegment, segment)
            if segment.type == "text":
                return await self.call_api(
                    "/message/custom/send",
                    touser=user_id,
                    msgtype="text",
                    text={"content": segment.data["text"]},
                )
            elif segment.type == "image":
                raise ApiNotAvailable("Image message is not supported in miniprogram")
            elif segment.type == "link":
                raise ApiNotAvailable("Link message is not supported in miniprogram")
            elif segment.type == "miniprogrampage":
                raise ApiNotAvailable("Miniprogram page message is not supported in miniprogram")



from typing import Any, Union, Callable, Optional, cast, Literal, Type, TYPE_CHECKING
from typing_extensions import override
from pydantic import BaseModel, Field

from nonebot.adapters import Event as BaseEvent
from nonebot.utils import logger_wrapper
from nonebot.compat import model_dump

from ..message import Message, MessageSegment


log = logger_wrapper("WXMP")


class Event(BaseEvent):
    to_user_id: str = Field(alias="ToUserName")
    """ 接收者的 OpenId `ToUserName` """
    user_id: str = Field(alias="FromUserName")
    """ 发送者的 OpenId `FromUserName` """
    timestamp: int = Field(alias="CreateTime")
    """ 消息发送时间戳 `CreateTime` """
    message_type: str = Field(alias="MsgType")
    """ 消息类型 `MsgType` """
    event: Optional[str] = Field(default=None, alias="Event")

    def handle_origin_data(self):
        """ 处理原始数据 """

    @override
    def get_type(self) -> str:
        return self.message_type

    @override
    def get_event_name(self) -> str:
        return self.__class__.__name__

    @override
    def get_event_description(self) -> str:
        return str(model_dump(self))

    @override
    def get_message(self) -> Message:
        raise NotImplementedError

    @override
    def get_plaintext(self) -> str:
        raise NotImplementedError

    @override
    def get_user_id(self) -> str:
        return str(self.user_id)

    @override
    def get_session_id(self) -> str:
        return f"{self.__class__.__name__}_{self.message_type}_{self.user_id}_{self.to_user_id}"

    @override
    def is_tome(self) -> bool:
        return True

    @classmethod
    def payload_to_event(cls, payload: dict) -> "Event":
        """ 将 payload 转换为 Event 对象的子类 """
        deepest_event = None
        deepest_depth = -1

        def dfs(current_cls: Type[Event], current_depth):
            nonlocal deepest_event, deepest_depth
            event = None
            for subclass in current_cls.__subclasses__():
                dfs(subclass, current_depth + 1)
            try:
                event = current_cls.model_validate(payload)
                event.handle_origin_data()
            except Exception:
                pass
            else:
                if current_depth > deepest_depth:
                    deepest_event = event
                    deepest_depth = current_depth

        dfs(cls, 0)
        if deepest_event:
            return deepest_event
        else:
            log("ERROR", f"Payload: {payload}")
            raise NotImplementedError


class MessageEvent(Event):
    """ 消息事件 """
    message_type: Literal["text", "image", "miniprogrampage"] = Field(alias="MsgType")
    """ 事件类型 """
    message_id: int = Field(alias="MsgId")
    """ 消息 ID `MsgId` """
    message: Optional[Message] = Field(default=None, alias="__message__")
    """ 消息内容 """

    @override
    def handle_origin_data(self) -> None:
        """ 处理消息内容 """
        if self.message_type == "text":
            self.message = Message(MessageSegment.text(text=getattr(self, "Content")))
        elif self.message_type == "image":
            self.message = Message(MessageSegment.image(file=getattr(self, "PicUrl")))
        elif self.message_type == "miniprogrampage":
            self.message = Message(MessageSegment.miniprogrampage(
                title=getattr(self, "Title"),
                appid=getattr(self, "AppId"),
                page_path=getattr(self, "PagePath"),
                thumb_media=getattr(self, "ThumbUrl"),
            ))
        else:
            raise NotImplementedError

    @override
    def get_message(self) -> Message:
        return Message(self.message)

    @override
    def get_plaintext(self) -> str:
        return "".join(segment.data.get("text", "") for segment in self.message)

    def get_type(self) -> str:
        return "message"



from typing import Any, Union, Callable, Optional, cast, Type
from typing_extensions import override
from pydantic import BaseModel, Field
from yarl import URL
import xmltodict
import asyncio
import hashlib
import json
import sys
import re

from nonebot import get_plugin_config
from nonebot.utils import logger_wrapper
from nonebot.exception import WebSocketClosed
from nonebot.utils import DataclassEncoder, escape_tag
from nonebot.drivers import (
    URL,
    Driver,
    Request,
    Response,
    WebSocket,
    ForwardDriver,
    ReverseDriver,
    HTTPServerSetup,
    WebSocketServerSetup,
)

from nonebot.adapters import Adapter as BaseAdapter

from .bot import Bot
from .event import *
from .config import Config, BotInfo
from .message import Message, MessageSegment


from nonebot import get_plugin_config
from nonebot.drivers import (
    Request,
    ASGIMixin,
    WebSocket,
    HTTPServerSetup,
    WebSocketServerSetup
)
from nonebot.exception import (
    ActionFailed,
    NetworkError,
    ApiNotAvailable,
)


log = logger_wrapper("WXMP")


class Adapter(BaseAdapter):
    @override
    def __init__(self, driver: Driver, **kwargs: Any):
        super().__init__(driver, **kwargs)
        self.wxmp_config: Config = get_plugin_config(Config)
        self.tasks: set["asyncio.Task"] = set()
        self.setup()

    @classmethod
    @override
    def get_name(cls) -> str:
        """ 适配器名称: `WXMP` """
        return "WXMP"

    def setup(self) -> None:
        if not isinstance(self.driver, ASGIMixin):
            raise RuntimeError(
                f"Current driver {self.config.driver} doesn't support asgi server!"
                f"{self.get_name()} Adapter need a asgi server driver to work."
            )

        for bot_info in self.wxmp_config.wxmp_bots:
            http_setup = HTTPServerSetup(
                URL(f"/wxmp/revice/{bot_info.appid}"),
                "POST",
                f"{self.get_name()} {bot_info.appid} Event",
                self._handle_event,
            )
            self.setup_http_server(http_setup)
            if not (bot := self.bots.get(bot_info.appid, None)):
                bot = Bot(self, bot_info.appid, bot_info)
                self.bot_connect(bot)
                log("INFO", f"<y>Bot {escape_tag(bot_info.appid)}</y> connected")

            if bot_info.enable_verify:
                http_setup = HTTPServerSetup(
                    URL(f"/wxmp/revice/{bot_info.appid}"),
                    "GET",
                    f"{self.get_name()} {bot_info.appid} Verify",
                    self._handle_verify,
                )
                self.setup_http_server(http_setup)

        http_setup = HTTPServerSetup(
            URL(f"/wxmp/revice"),
            "GET",
            f"{self.get_name()} Root Verify",
            self._handle_verify,
        )
        self.setup_http_server(http_setup)

        http_setup = HTTPServerSetup(
            URL(f"/wxmp/revice"),
            "POST",
            f"{self.get_name()} Root Event",
            self._handle_event,
        )
        self.setup_http_server(http_setup)

    @classmethod
    def parse_body(cls, data: str) -> dict:
        try:
            return json.loads(data)
        except json.JSONDecodeError:
            res: dict = xmltodict.parse(data)
            if _res := res.get("xml", None):
                return _res
            else:
                return res

    async def _handle_event(self, request: Request) -> Response:
        """ 处理微信公众平台的事件推送 """
        url = URL(request.url)
        timestamp = url.query.get("timestamp", "")
        nonce = url.query.get("nonce", "")
        signature = url.query.get("signature", "")

        bot: Bot = self.bots.get(self._get_appid(url.path), None)

        if not bot:
            return Response(200, content="success")

        if request.content:
            concat_string: str = ''.join(sorted([bot.bot_info.token, timestamp, nonce]))
            sha1_signature = hashlib.sha1(concat_string.encode('utf-8')).hexdigest()
            if sha1_signature != signature:
                return Response(403, content="Invalid signature")
            else:
                payload: dict = self.parse_body(request.content)

                if bot.bot_info.type == "miniprogram":  # 小程序
                    event = MiniprogramEvent.payload_to_event(payload)
                elif bot.bot_info.type == "official":  # 公众号
                    event = OfficalAccountEvent.payload_to_event(payload)
                else:
                    raise ValueError(f"Invalid bot type: {bot.bot_info.type}")

                bot = self.bots.get(bot.bot_info.appid)
                task = asyncio.create_task(bot.handle_event(event))
                task.add_done_callback(self.tasks.discard)
                self.tasks.add(task)

                return Response(200, content="success")
        else:
            return Response(400, content="Invalid request body")

    async def _handle_verify(self, request: Request) -> Any:
        """ 响应微信公众平台的签名验证 """
        url = URL(request.url)
        signature = url.query.get("signature", "")
        echostr = url.query.get("echostr", "")
        timestamp = url.query.get("timestamp", "")
        nonce = url.query.get("nonce", "")

        bot: Bot = self.bots.get(self._get_appid(url.path), None)

        if not bot:  # 默认验证通过
            return Response(200, content=echostr)

        concat_string: str = ''.join(sorted([timestamp, nonce, bot.bot_info.token]))
        sha1_signature = hashlib.sha1(concat_string.encode('utf-8')).hexdigest()

        if sha1_signature == signature:
            return Response(200, content=echostr)
        else:
            return Response(403, content="Invalid signature")

    def _get_appid(self, path: str) -> BotInfo | None:
        """ 从链接中获取 Bot 配置 """
        return path.split('/')[-1]

    async def _call_api(self, bot: Bot, api: str, **data: Any) -> Any:
        """ 调用微信公众平台 API """
        access_token = await bot._get_access_token()
        request = Request(
            method="POST",
            url=f"https://api.weixin.qq.com/cgi-bin{api}",
            params={
                "access_token": access_token,
            },
            headers={
                'Content-type': 'application/json',
            },
            content=json.dumps(data, ensure_ascii=False).encode("utf-8"),
        )
        resp = await self.request(request)
        if resp.status_code != 200 or not resp.content:
            raise NetworkError(f"Call API {api} failed with status code {resp.status_code}.")
        res: dict = json.loads(resp.content)
        if res.get("errcode", 0) != 0:
            log("ERROR", f"Call API {api} failed with error {res}")
            raise ActionFailed()
        return res



from .base import *
from .offical import *
from .miniprogram import *

__all__ = [
    "Event",
    "MessageEvent",
    "MiniprogramEvent",
    "OfficalAccountEvent",

    "AuthorizationChangeEvent",
    "MiniprogramMessageEvent",
    "OfficalAccountMessageEvent",
    "SubscribeEvent",
    "UnSubscribeEvent",
    "MenuClickEvent",
]



from typing import Any, Union, Callable, Optional, cast, Literal, Type
from pydantic import Field

from .base import Event, MessageEvent


class MiniprogramEvent(Event):
    """ 小程序 事件 """


class MiniprogramMessageEvent(MessageEvent, MiniprogramEvent):
    """ 小程序 客服消息事件 """


class UserEnterEvent(MiniprogramEvent):
    """ 用户进入客服会话事件 """
    message_type: Literal["event"] = Field(alias="MsgType")
    event: Literal["user_enter_tempsession"] = Field(alias="Event")
    session_from: str = Field(alias="SessionFrom")
    """ 会话来源，开发者在客服会话按钮设置的 session-from 属性 """


class AuthorizationChangeEvent(MiniprogramEvent):
    """ 授权用户信息变更事件 """
    message_type: Literal["event"] = Field(alias="MsgType")
    """ 消息类型 `MsgType` """
    event: Literal["user_authorization_revoke"] = Field(alias="Event")
    """ 事件类型 `Event` """
    openid: str = Field(alias="OpenID")
    """ 用户 OpenID `OpenID` """
    appid: str = Field(alias="AppID")
    """ 小程序 AppID `AppID` """
    revoke_info: str = Field(alias="RevokeInfo")
    """ 取消授权的数据类型 `RevokeInfo` """



from typing import Type, Union, Mapping, Iterable, Optional, Self
from typing_extensions import override

from nonebot.adapters import Message as BaseMessage, MessageSegment as BaseMessageSegment


class MessageSegment(BaseMessageSegment["Message"]):
    """ 消息段 """

    @classmethod
    @override
    def get_message_class(cls) -> Type["Message"]:
        return Message

    @override
    def __str__(self) -> str:
        return self.data["text"]

    @override
    def is_text(self) -> bool:
        return self.type == "text"

    @classmethod
    def text(cls, text: str) -> Self:
        """ 文本消息段 """
        return cls("text", {"text": text})

    @classmethod
    def image(cls, file: Union[bytes, str]) -> Self:
        """ 图片消息段 """
        return cls("image", {
            "file": file,
        })

    @classmethod
    def link(cls, title: str, description: str, url: str, thumb_url: str) -> Self:
        """ 链接消息段 """
        return cls("link", {
            "title": title,
            "description": description,
            "url": url,
            "thumb_url": thumb_url,
        })

    @classmethod
    def miniprogrampage(cls, title: str, appid: str, page_path: str, thumb_media: str) -> Self:
        """ 小程序卡片消息段 """
        return cls("miniprogrampage", {
            "title": title,
            "appid": appid,
            "page_path": page_path,
            "thumb_media": thumb_media,
        })

    def __iter__(self):
        """ """
        yield self


class Message(BaseMessage[MessageSegment]):
    """ 消息 """

    @classmethod
    @override
    def get_segment_class(cls) -> Type[MessageSegment]:
        return MessageSegment

    @staticmethod
    @override
    def _construct(msg: str) -> Iterable[MessageSegment]:
        return MessageSegment.text(msg)
